/*
 * LetterOrder - Experimenting with the order of letters in words
 * Copyright (C) 2011 Christian Schenk
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 */
package org.christianschenk.letterorder.utils;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.christianschenk.letterorder.utils.enums.WordDistance;

import com.google.inject.Inject;

/**
 * Uses permutations or pure chance to order letters in words. Will make sure
 * that the first and last character of a word is kept as is.
 * 
 * @author Christian Schenk
 */
public class LetterOrderUtils {

	private final PermutationUtils permutationUtils;
	private final StringDistanceUtils stringDistanceUtils;

	@Inject
	public LetterOrderUtils(final PermutationUtils permutationUtils, final StringDistanceUtils stringDistanceUtils) {
		this.permutationUtils = permutationUtils;
		this.stringDistanceUtils = stringDistanceUtils;
	}

	/**
	 * Returns the given list of words in an obfuscated form. Depending on the
	 * given {@link WordDistance} the words are either mostly easy or really
	 * hard to read.
	 * 
	 * @param words
	 *            list of words
	 * @param wordDistance
	 *            distance between original and obfuscated word
	 * @return list of obfuscated words
	 */
	public List<String> getObfuscatedWords(final List<String> words, final WordDistance wordDistance) {
		final List<String> obfuscatedWords = new ArrayList<String>();

		for (final String word : words) {
			// 1. get obfuscated words
			final List<String> obfuscatedWordVariants = this.getObfuscatedWordVariants(word);

			// 2. calculate distance
			// 3. select word based on distance
			final String selectedWord = this.getWordBasedOnDistance(word, obfuscatedWordVariants, wordDistance);

			// 4. store in result set
			obfuscatedWords.add(selectedWord);
		}

		return obfuscatedWords;
	}

	/**
	 * Returns a list of variants of the given word either using permutations or
	 * simply a random order of the letters of the word.
	 * 
	 * @param word
	 *            a word
	 * @return list of variants of the given word
	 */
	private List<String> getObfuscatedWordVariants(final String word) {
		final List<String> words = new ArrayList<String>();

		final int relevantLetters = word.length() - 2;
		// The word is smaller than 4 characters thus we don't change it
		if (relevantLetters < 2) {
			words.add(word);
		} else {
			final int[][] permutations;
			// If the relevant letters in the word are larger than 8 characters
			// using permutations might not be possible due to performance
			// issues. In the latter case we will use a simple random order.
			if (relevantLetters < 8) {
				permutations = this.permutationUtils.getPermutations(relevantLetters);
			} else {
				permutations = new int[1][];
				permutations[0] = this.permutationUtils.getRandomPermutation(relevantLetters);
			}

			final String workLetters = word.substring(1, word.length() - 1);
			for (final int[] permutation : permutations) {
				final String rearrangedLetters = this.getRearrangedLetters(workLetters, permutation);
				final String rearrangedWord = word.charAt(0) + rearrangedLetters + word.charAt(word.length() - 1);
				if (words.contains(rearrangedWord) == false) {
					words.add(rearrangedWord);
				}
			}
		}

		return words;
	}

	/**
	 * Rearranges the letters based on the given pattern. Silently assumes that
	 * the lenght of <code>letters</code> is equal to the lenght of
	 * <code>pattern</code>.
	 * 
	 * @param letters
	 *            string with certain letters
	 * @param pattern
	 *            list of indices that define the order of the letters
	 * @return reordered letters based on the given pattern
	 */
	private String getRearrangedLetters(final String letters, final int[] pattern) {
		final char[] rearranged = new char[letters.length()];
		for (int i = 0, n = pattern.length; i < n; i++) {
			rearranged[i] = letters.charAt(pattern[i]);
		}
		return new String(rearranged);
	}

	/**
	 * Returns a word from the given list of words according to the given
	 * distance. In case there are more words with the same distance a word is
	 * randomly selected.
	 * 
	 * @param originalWord
	 *            the original word to calculate the distance
	 * @param words
	 *            list of words to select from
	 * @param wordDistance
	 *            preferred distance
	 * @return selected word
	 */
	@SuppressWarnings("boxing")
	private String getWordBasedOnDistance(final String originalWord, final List<String> words, final WordDistance wordDistance) {
		// Calculate distances for all words
		final Map<Integer, List<String>> distances = new HashMap<Integer, List<String>>();
		for (final String word : words) {
			final int distance = this.stringDistanceUtils.getLevenshteinDistance(originalWord, word);
			if (distance == 0) continue;
			List<String> distance2Words = distances.get(distance);
			if (distance2Words == null) {
				distance2Words = new ArrayList<String>();
				distances.put(distance, distance2Words);
			}
			distance2Words.add(word);
		}

		// Select one randomly
		if (distances.isEmpty()) {
			return words.get(RandomUtils.RANDOM.nextInt(words.size()));
		}

		// Select word based on given distance
		final List<Integer> keys = new ArrayList<Integer>(distances.keySet());
		final List<String> selectedWords = wordDistance.equals(WordDistance.LOW) ? distances.get(keys.get(0)) : distances.get(keys.get(keys.size() - 1));
		final String selectedWord = selectedWords.get(RandomUtils.RANDOM.nextInt(selectedWords.size()));
		return selectedWord;
	}
}